"""
The BCOM Index is defined in https://assets.bbhub.io/professional/sites/10/BCOM-Methodology.pdf
This model here shows a basic implementation of these rules, not including any logic to handle
trading disruption events, etc.
"""
from datetime import date, datetime
from typing import Mapping

import polars as pl
from frozendict import frozendict as fd
from hg_oap.instruments.future import month_from_code
from hgraph import graph, TS, const, TSD, index_of, if_then_else, switch_, lift, CmpResult, explode, register_service, \
    default_path, cast_, Frame, map_
from importlib_resources import files, as_file

from hg_systematic.impl import calendar_for_static, StaticPriceSchema
from hg_systematic.impl._calendar_impl import create_market_holidays
from hg_systematic.operators import index_composition, index_rolling_contracts, \
    index_rolling_weight, business_days, Periods, \
    HolidayCalendar, symbol_is, INDEX_ROLL_FLOAT, INDEX_ROLL_STR
from hg_systematic.operators._rolling_rules import monthly_rolling_info, MonthlyRollingRequest, \
    rolling_contracts, bbg_commodity_contract_fn


@graph(overloads=index_rolling_weight, requires=symbol_is("BCOM Index"))
def index_rolling_weights_bcom(symbol: str, dt: TS[date], calendar: HolidayCalendar) -> TS[float]:
    days_of_month = business_days(Periods.Month, calendar, dt)
    day_index = index_of(days_of_month, dt) + 1
    is_rolling = if_then_else(day_index < 6, CmpResult.LT, if_then_else(day_index > 9, CmpResult.GT, CmpResult.EQ))
    return switch_(
        is_rolling,
        {
            CmpResult.LT: lambda d: const(1.0),
            CmpResult.EQ: lambda d: lift(lambda x: round(x, 1), inputs={"x": TS[float]}, output=TS[float])(
                (10 - d) * .2),
            CmpResult.GT: lambda d: const(0.0),
        },
        cast_(float, day_index),
    )


def get_cims_for_year(year: int) -> Mapping[str, float]:
    """Load the cims from the csv file"""
    import examples.bcom_index
    source = files(examples.bcom_index).joinpath("bcom_cims.csv")
    with as_file(source) as resource_path:
        df = pl.read_csv(resource_path)
    year_column = str(year)
    return fd(df.select("Commodity", year_column).iter_rows())


@graph(overloads=index_composition, requires=symbol_is("BCOM Index"))
def index_composition_bcom(symbol: str, dt: TS, calendar: HolidayCalendar) -> INDEX_ROLL_FLOAT:
    y, m, _ = explode(dt)
    y_prev = y - 1
    get_cims = lift(get_cims_for_year, output=TSD[str, TS[float]])
    cim2 = get_cims(y)
    prev = get_cims(y_prev)
    cim1 = if_then_else(m == 1, prev, cim2)
    return INDEX_ROLL_FLOAT.from_ts(first=cim1, second=cim2)


def get_bcom_roll_schedule() -> Mapping[str, Mapping[int, tuple[int, int]]]:
    import examples.bcom_index
    source = files(examples.bcom_index).joinpath("bcom_roll_schedule.csv")
    with as_file(source) as resource_path:
        df = pl.read_csv(resource_path)
    return fd((k[0], fd(
        (month_from_code(k_),
         (month_from_code((i := v_.item())[0]), int(i[1]))) for k_, v_ in v.to_dict().items())) for
              k, v in df.partition_by("Commodity", include_key=False, as_dict=True).items())


@graph(overloads=index_rolling_contracts, requires=symbol_is("BCOM Index"))
def index_rolling_contracts_bcom(symbol: str, dt: TS, calendar: HolidayCalendar) -> INDEX_ROLL_STR:
    rs = get_bcom_roll_schedule()
    roll_schedule = const(rs, TSD[str, TSD[int, TS[tuple[int, int]]]])
    fmt_str = const(fd({k: f"{k}{{month}}{{year:02d}} Comdty" for k in rs.keys()}), TSD[str, TS[str]])
    year_scale = const(fd({k: 100 for k in rs.keys()}), TSD[str, TS[int]])
    rolling_info = monthly_rolling_info(
        request=MonthlyRollingRequest(5, 10, "BCOM")
    )
    contracts = map_(
        lambda asset, ri, rs, c_fn: rolling_contracts(ri, rs, asset, c_fn),
        rolling_info,
        roll_schedule,
        bbg_commodity_contract_fn,
        __keys__=roll_schedule.key_set,
        __key_arg__="asset",
    )
    first = map_(lambda c: c[0], contracts)
    second = map_(lambda c: c[1], contracts)
    return INDEX_ROLL_STR.from_ts(first=first, second=second)


def create_bcom_holidays() -> frozenset[date]:
    return create_market_holidays(["US", "GB"], datetime(1996, 1, 1), datetime(2035, 1, 1))


def register_bcom_static_calendar():
    """
    Use an approximation of the holidays using a US and GB calendar. For now load from 1996 through 2035.
    """
    register_service(default_path, calendar_for_static, holidays=fd(
        {"BCOM Index": create_bcom_holidays()}))


def load_sample_prices() -> Frame[StaticPriceSchema]:
    import examples.bcom_index
    source = files(examples.bcom_index).joinpath("bcom_prices.csv")
    with as_file(source) as resource_path:
        df = pl.read_csv(resource_path)
    return df.melt("Commodity", variable_name="date", value_name="price").cast({"date": date}).rename(
        {"Commodity": "symbol"}).select("date", "symbol", "price").sort("date")



from dataclasses import dataclass
from datetime import date
from typing import Callable

from hg_oap.instruments.future import month_code
from hgraph import TimeSeriesSchema, TS, subscription_service, default_path, CompoundScalar, graph, TSD, TSB, TSL, Size, \
    CmpResult, reference_service, format_, lift, compute_node, apply

__all__ = ["MonthlyRollingRange", "monthly_rolling_weights", "MonthlyRollingRequest", "MonthlyRollingWeightRequest",
           "monthly_rolling_info", "MonthlyRollingInfo", "rolling_contracts", "bbg_commodity_contract_fn",
           "rolling_schedules"]


@dataclass(frozen=True)
class MonthlyRollingRequest(CompoundScalar):
    """
    Defines the rolling request parameters.
    This is the base information needed to determine a monthly roll schedule.

    :param start: The start of the roll range, this can be negative.
    :param end: The end of the roll range. This may not be negative.
    :param calendar_name: The holiday calendar to use when determining business dates.
    """
    start: int
    end: int
    calendar_name: str


@dataclass(frozen=True)
class MonthlyRollingWeightRequest(MonthlyRollingRequest):
    """
    Specified a linear roll over the range specified.
    The start can be negatively offset to indicate the roll to this month's contract
    starts in the prior month.
    The start and end date may never overlap, there MUST be the opportunity for at
    least one value of 1.0 and one of 0.0 in any month.
    """
    round_to: int


@subscription_service
def monthly_rolling_weights(request: TS[MonthlyRollingWeightRequest], path: str = default_path) -> TS[float]:
    """
    Produces a stream of rolling weights over the given calendars business days.
    This will only tick a value if the result is modified, i.e. it does not tick
    each time a date changes, but only when the result is different.
    """


@dataclass
class MonthlyRollingRange(TimeSeriesSchema):
    start: TS[int]
    end: TS[int]
    first_day: TS[int]  # This is the same as start when start is a positive value, and is the day index of the
    # previous month when negative


@dataclass
class MonthlyRollingInfo(MonthlyRollingRange):
    days_of_month: TS[tuple[date]]  # The dates within the month
    day_index: TS[int]  # The index within the days_of_month
    dt: TS[date]  # The date that the information represents
    # --- exploded values of date
    day: TS[int]
    month: TS[int]
    year: TS[int]
    # ---
    begin_roll: TS[bool]
    end_roll: TS[bool]
    roll_state: TS[CmpResult]  # LT before roll, EQ in roll, GT After roll
    roll_out_month: TS[int]
    roll_out_year: TS[int]
    roll_in_month: TS[int]
    roll_in_year: TS[int]


@subscription_service
def monthly_rolling_info(request: TS[MonthlyRollingRequest], path: str = default_path) -> TSB[MonthlyRollingInfo]:
    """
    The raw rolling information that can used to generate out weights or contracts.
    """


@graph
def rolling_contracts(
        roll_info: TSB[MonthlyRollingInfo],
        roll_schedule: TSD[int, TS[tuple[int, int]]],
        asset: TS[str],
        contract_fn: TS[Callable[[TS[str], TS[int], TS[int]], TS[str]]],
) -> TSL[TS[str], Size[2]]:
    """
    The contracts for the given roll_info and contract_fn.

    The ``contract_fn`` converts the month and year into a string for the name.
    For example:

    ::

        @graph
        def bbg_commodity_contract_fn(asset: TS[str], month: TS[str], year: TS[int]) -> TS[str]:
            y = year % 100
            return format_(
                "{asset}{month}{year} Comdty",
                month=lift(month_code, inputs={"d": TS[int]})(month),
                year=y
            )

    """
    m1 = roll_info.roll_out_month
    y1 = roll_info.roll_out_year
    m2 = roll_info.roll_in_month
    y2 = roll_info.roll_in_year

    c1_m = _create_contract(
        month=m1,
        year=y1,
        schedule=roll_schedule,
        asset=asset,
        contract_fn=contract_fn
    )

    c2_m = _create_contract(
        month=m2,
        year=y2,
        schedule=roll_schedule,
        asset=asset,
        contract_fn=contract_fn
    )

    return TSL.from_ts(c1_m, c2_m)


@graph
def _create_contract(
        month: TS[int],
        year: TS[int],
        schedule: TSD[int, TS[tuple[int, int]]],
        asset: TS[str],
        contract_fn: TS[Callable[[TS[str], TS[int], TS[int]], TS[str]]]) -> TS[str]:
    s = schedule[month]
    m = s[0]
    y = year + s[1]
    return apply[TS[str]](contract_fn, asset, m, y)


@reference_service
def rolling_schedules(path: str = default_path) -> TSD[str, TSD[int, TS[tuple[int, int]]]]:
    """
    The rolling contracts for the symbol.
    """


def bbg_commodity_contract_fn(asset: str, month: int, year: int) -> str:
    return f"{asset}{month_code(month)}{year % 100:02d} Comdty"



from typing import Mapping
from frozendict import frozendict as fd
from hg_oap.instruments.future import month_from_code
from hgraph import lift, TSD, TS

__all__ = ["roll_schedule_to_map", "roll_schedule_to_tsd"]

def roll_schedule_to_map(roll_schedule: tuple[str, ...]) -> Mapping[int, tuple[int, int]]:
    """convert a roll schedule tuple to a mapping"""
    assert all(item[1] in ('0', '1') for item in roll_schedule), "Roll year must be 0 or 1"
    return fd({ndx+1: (month_from_code(item[0]), int(item[1])) for ndx, item in enumerate(roll_schedule)})


roll_schedule_to_tsd = lift(roll_schedule_to_map, output=TSD[int, TS[tuple[int, int]]])



import pandas as pd
import yfinance as yf
from datetime import date


def get_bcom_cl_contracts(start_year, end_year):
    """
    Grab the contracts used in pricing the BCOM index over the range
    specified.
    """
    all_contracts = []
    for year in range(start_year, end_year + 1):
        # Define contract month codes: Jan=F, Feb=G, ..., Dec=Z
        months = ['F', 'G', 'H', 'J', 'K', 'M', 'N', 'Q', 'U', 'V', 'X', 'Z']

        for month in months:
            contract = f"CL{month}{str(year)[-2:]}"
            all_contracts.append(contract)

    return all_contracts


# Step 2: This only seems to work with contracts that are current live
# So we can't get historical data really.
bcom_roll_schedule = get_bcom_cl_contracts(y := date.today().year, y+3)
print(bcom_roll_schedule)


# Step 3: Optional - Download Data using API like Yahoo Finance (example for CL=F)
# This part would fetch historical prices for each contract in the rolling schedule if necessary
def download_cl_prices(contracts):
    data = {}
    for contract in contracts:
        try:
            # Download each contract's price data from Yahoo Finance
            print(f"Downloading prices for {contract}...")
            df = yf.download(f"{contract}.NYM")
            if df is not None and len(df) > 0:
                data[contract] = df['Close']  # Use Close prices
        except Exception as e:
            print(f"Error in downloading {contract}: {str(e)}")
    return pd.concat(data, join="outer", axis=1)

# Example usage (prices download is optional and requires valid contract symbols):
# cl_prices = download_cl_prices(bcom_roll_schedule)



from typing import Mapping

from hgraph import subscription_service, TSS, TS, service_impl, const, TSD, map_

from hg_systematic.index.configuration import IndexConfiguration

__all__ = ["index_configuration", "static_index_configuration", ]


@subscription_service
def index_configuration(symbol: TS[str]) -> TS[IndexConfiguration]:
    """
    Retrieves the definition of the desired index.
    """


@service_impl(interfaces=index_configuration)
def static_index_configuration(symbol: TSS[str], indices: Mapping[str, IndexConfiguration]) \
        -> TSD[str, TS[IndexConfiguration]]:
    """
    Provide a simple implementation of the index configuration service that supports a manual mapping of symbol to
    configurations.
    """
    indices = const(indices, TSD[str, TS[IndexConfiguration]])
    return map_(lambda config: config, indices, __keys__=symbol)



from dataclasses import dataclass

from hgraph import subscription_service, TSS, TS, TSD, mesh_, graph, service_impl, dispatch_, dispatch, operator, \
    TimeSeriesSchema, TSB, default_path

from hg_systematic.index.configuration import IndexConfiguration
from hg_systematic.index.configuration_service import index_configuration
from hg_systematic.index.units import IndexStructure

__all__ = ["price_index", "price_index_impl", "INDEX_MESH", "IndexResult", "price_index_op"]


@dataclass
class IndexResult(TimeSeriesSchema):
    level: TS[float]
    index_structure: TSB[IndexStructure]


@subscription_service
def price_index(symbol: TS[str], path: str=default_path) -> TSB[IndexResult]:
    """
    Produce a price for an index.
    """


INDEX_MESH = "index_mesh"


@service_impl(interfaces=price_index)
def price_index_impl(symbol: TSS[str]) -> TSD[str, TSB[IndexResult]]:
    """
    The basic structure for implementing the index pricing service. This makes use of the mesh_ operator allowing
    for nested pricing structures.
    """
    return _price_index_mesh(symbol)


@graph
def _price_index_mesh(symbol: TSS[str]) -> TSD[str, TSB[IndexResult]]:
    """Separate the mesh impl to make testing easier."""
    return mesh_(
        _price_index,
        __keys__=symbol,
        __key_arg__="symbol",
        __name__=INDEX_MESH
    )


@graph
def _price_index(symbol: TS[str]) -> TSB[IndexResult]:
    """Loads the index configuration object and dispatches it"""
    config = index_configuration(symbol)
    return price_index_op(config)


@dispatch(on=("config",))
@operator
def price_index_op(config: TS[IndexConfiguration]) -> TSB[IndexResult]:
    """
    Dispatches to the appropriate pricing implementation based on the configuration instance.
    To implement an index, implement the price_index_op operator.
    """



from typing import TypeVar, Callable

from frozendict import frozendict
from hgraph import graph, TSB, TS, map_, reduce, dedup, or_, and_, len_, DebugContext, combine, switch_, TS_SCHEMA, \
    sample, default, gate, not_, if_then_else, CmpResult, no_key, const, AUTO_RESOLVE, TS_SCHEMA_1, feedback, lag, \
    contains_, round_

from hg_systematic.index.configuration import IndexConfiguration, initial_structure_from_config
from hg_systematic.index.pricing_service import IndexResult
from hg_systematic.index.units import IndexPosition, NotionalUnitValues, IndexStructure, NotionalUnits
from hg_systematic.operators import MonthlyRollingInfo, monthly_rolling_info, monthly_rolling_weights, \
    MonthlyRollingWeightRequest, calendar_for

ROLLING_CONFIG = TypeVar("ROLLING_CONFIG", bound=IndexConfiguration)


@graph
def monthly_rolling_index(
        config: TS[ROLLING_CONFIG],
        prices: NotionalUnitValues,
        compute_target_units_fn: Callable[[TSB[TS_SCHEMA]], NotionalUnitValues],
        re_balance_signal_fn: Callable[[TSB[TS_SCHEMA]], TS[bool]] = None,
        _config_tp: type[ROLLING_CONFIG] = AUTO_RESOLVE,
        _kwargs_tp: type[TS_SCHEMA_1] = AUTO_RESOLVE,
        **kwargs: TSB[TS_SCHEMA_1]
) -> TSB[IndexResult]:
    """
    Provide the wrapper for wiring in the required data sources for the monthly_rolling_index_component.
    This makes the assumptions about the data sources that are required and sets up the feedback loops required.

    This item itself cannot be recorded, however, it will call the monthly_rolling_index_component which should be
    recordable. Ultimately this component should be recordable and point-in-time recoverable.

    If a custom halt_signal is provided (through the kwargs) this is used, alternatively, if the config contains the
    trading_halt_calendar then this is used to load the specified calendar. If neither of these are provided then
    the default of False (i.e. never halted) is used.

    Roll info and weights are obtained using the get_monthly_rolling_values function, using the config to extract
    configuration to retrieve the data, unless supplied as part of the kwargs

    :param config: The configuration for this index.
    :param prices: The current price of the contracts or the levels of the sub-indices.
    :param compute_target_units_fn: A function that computes the target units for re-balancing.
    :param re_balance_signal_fn: A function that returns a signal to filter the monthly re-balance signal.
    :param _config_tp: The type of the config, will AUTO_RESOLVE not to be supplied.
    :param _kwargs_tp: The type of the kwargs, will AUTO_RESOLVE not to be supplied.
    """
    cfg_schema = _config_tp.__meta_data_schema__
    kwargs_schema = _kwargs_tp.__meta_data_schema__

    if "roll_info" in kwargs_schema:
        if "roll_weight" not in kwargs_schema:
            raise ValueError("roll_info is provided, but roll_weight is not.")
        roll_info = kwargs["roll_info"]
        roll_weight = kwargs["roll_weight"]
    else:
        roll_info, roll_weight = get_monthly_rolling_values(config)

    # If a custom halt_trading signal is provide, use it, alternatively if the config has a trading_halt_calendar
    # use that to feed into a simple contains filter, finally if no other strategy is provided assume no halt signal
    # and wire in False.
    halt_trading = kwargs.halt_trading if "halt_trading" in kwargs_schema else \
        dedup(contains_(calendar_for(config.trading_halt_calendar), roll_info.dt)) if "trading_halt_calendar" in cfg_schema else \
                  const(False)
    DebugContext.print("halt_trading", halt_trading)

    if re_balance_signal_fn is None:
        if "re_balance_month" in cfg_schema:
            re_balance_signal_fn = lambda tsb: tsb.config.re_balance_month == tsb.roll_info.roll_out_month
        else:
            # Defaults to True
            re_balance_signal_fn = lambda tsb: const(True)

    index_structure_fb = feedback(TSB[IndexStructure])
    DebugContext.print("index_structure_fb", index_structure_fb())
    index_structure = dedup(default(lag(index_structure_fb(), 1, roll_info.dt), initial_structure_from_config(config)))
    DebugContext.print("index_structure", index_structure)

    out = monthly_rolling_index_component(
        config,
        index_structure,
        roll_weight,
        roll_info,
        prices,
        halt_trading,
        re_balance_signal_fn=re_balance_signal_fn,
        compute_target_units_fn=compute_target_units_fn,
        **{k: kwargs[k] for k in kwargs_schema if k not in {"halt_trading", "roll_info", "roll_weight"}}
    )

    # There is a dedup here as there seems to be a bug somewhere when dealing with REFs and TSD, will trace down later.
    index_structure_fb(dedup(out.index_structure))  # TODO: Find out why not using de-dup causes this to fail.

    result = out.copy_with(level=round_(out.level, config.rounding))
    DebugContext.print("published level", result.level)
    return result


@graph
def monthly_rolling_index_component(
        config: TS[ROLLING_CONFIG],
        index_structure: TSB[IndexStructure],
        roll_weight: TS[float],
        roll_info: TSB[MonthlyRollingInfo],
        prices: NotionalUnitValues,
        halt_trading: TS[bool],
        re_balance_signal_fn: Callable[[TSB[TS_SCHEMA]], TS[bool]],
        compute_target_units_fn: Callable[[TSB[TS_SCHEMA]], NotionalUnitValues],
        **kwargs: TSB[TS_SCHEMA]
) -> TSB[IndexResult]:
    """
    :param config: The configuration for this index.
    :param index_structure: The current index structure.
    :param roll_weight: The weight to transition from previous to current position.
    :param roll_info: The rolling information for this index.
    :param prices: The current price of the contracts of interest
    :param halt_trading: A signal to indicate that trading should be halted.
    :param re_balance_signal_fn: A function that returns a signal to trigger re-balancing.
    :param compute_target_units_fn: A function that computes the target units for re-balancing.
    :return: The level and other interim information.
    """

    # If we have already traded this produces an unnecessary computation, but check if we traded again
    # may be just as expensive and there is less switching involved then.
    level = compute_level(index_structure.current_position, prices)

    new_index_structure = re_balance_index(
        config=config,
        index_structure=index_structure,
        roll_info=roll_info,
        roll_weight=roll_weight,
        prices=prices,
        level=level,
        trade_halt=halt_trading,
        # We re-balance every time the begin_roll is triggered irrespective of trade-halt conditions.
        re_balance_signal_fn=re_balance_signal_fn,
        compute_target_units_fn=compute_target_units_fn,
        **kwargs
    )

    return combine[TSB[IndexResult]](
        level=level,
        index_structure=new_index_structure
    )


@graph
def compute_level(
        current_position: TSB[IndexPosition],
        current_value: NotionalUnitValues
) -> TS[float]:
    """
    Compute the level from the current positions and the last re-balance level
    """
    DebugContext.print("[compute_level] current_positions", current_position)
    DebugContext.print("[compute_level] current_value", current_value)
    returns = map_(
        lambda pos_curr, prc_prev, prc_now: (prc_now - prc_prev) * pos_curr,
        current_position.units,
        current_position.unit_values,
        current_value,
        __keys__=current_position.units.key_set,
    )
    DebugContext.print("[compute_level] returns", returns)
    new_level = current_position.level + reduce(
        lambda x, y: x + y,
        returns,
        0.0
    )
    DebugContext.print("[compute_level] level", new_level)
    return new_level


@graph
def needs_re_balance(
        index_structure: TSB[IndexStructure],
        rolling_info: TSB[MonthlyRollingInfo],
        re_balance_signal: TS[bool] = True
) -> TS[bool]:
    """
    Determines if the index needs to be re-balanced. A filter condition can be supplied to restrict
    when the re-balance is actually triggered, by default the value is set to True.
    This will turn True when the rolling info begin_roll is set to True, and will remain true until
    the target_units property of the index structure is set back to an empty set.

    The re_balance_signal is defaulted to True. If set, this must be True when the begin_roll is True in
    order to trigger the re-balance.
    """
    re_balance = dedup(or_(
        # This will initiate a roll, so will set the target units
        and_(rolling_info.as_schema.begin_roll, re_balance_signal),
        # Once the roll is complete, the target units are set to an empty dict.
        len_(index_structure.target_units) > 0,
    ))
    DebugContext.print("[needs_re_balance]", re_balance)
    return re_balance


# Check that the type includes the roll_period and roll_rounding values to be safe
@graph(requires=lambda m, s: {"roll_period", "roll_rounding"}.issubset(m[ROLLING_CONFIG].meta_data_schema))
def get_monthly_rolling_values(config: TS[ROLLING_CONFIG]) -> TSB["roll_info": TSB[MonthlyRollingInfo],
                                                              "weights": TS[float]]:
    monthly_rolling_request = combine[TS[MonthlyRollingWeightRequest]](
        start=config.roll_period[0],
        end=config.roll_period[1],
        calendar_name=config.publish_holiday_calendar,
        round_to=config.roll_rounding
    )
    DebugContext.print("[monthly_rolling] request", monthly_rolling_request)
    roll_info = monthly_rolling_info(monthly_rolling_request)
    DebugContext.print("[monthly_rolling] roll_info", roll_info)
    rolling_weights = monthly_rolling_weights(monthly_rolling_request)
    DebugContext.print("[monthly_rolling] rolling_weights", rolling_weights)
    return combine[TSB](roll_info=roll_info, weights=rolling_weights)


@graph
def re_balance_index(
        config: TS[ROLLING_CONFIG],
        index_structure: TSB[IndexStructure],
        roll_info: TSB[MonthlyRollingInfo],
        roll_weight: TS[float],
        prices: NotionalUnitValues,  # Uses prices instead of values to avoid conflict with values() method
        level: TS[float],
        trade_halt: TS[bool],
        re_balance_signal_fn: Callable[[TSB[TS_SCHEMA]], TS[bool]],
        compute_target_units_fn: Callable[[TSB[TS_SCHEMA]], NotionalUnitValues],
        **kwargs: TSB[TS_SCHEMA]
) -> TSB[IndexStructure]:
    """
    Re-balance the index when needs re-balance is True.
    Will call the re-balance function with the schema:
        * config
        * index_structure
        * roll_info
        * roll_weights
        * re_balance_signal
        * prices
        * level
        * trade_halt
        * **kwargs (expansion of any additional arguments provided)

    The function should be a graph or compute node that takes the form of:

    ::
        @graph
        def my_compute_target_units(tsb: TSB[TS_SCHEMA]) -> TSB[IndexStructure]:
            ...


    """
    DebugContext.print("[re_balance_index] index_structure:pre", index_structure)

    re_balance_signal = re_balance_signal_fn(combine[TSB](
        config=config,
        index_structure=index_structure,
        roll_info=roll_info,
        roll_weight=roll_weight,
        prices=prices,
        level=level,
        trade_halt=trade_halt,
        **kwargs,
    ))

    new_index_structure = switch_(
        needs_re_balance(index_structure, roll_info, re_balance_signal),
        {
            True: lambda tsb: _re_balance(tsb, compute_target_units_fn),
            False: _pass_through,
        },
        combine[TSB](
            config=config,
            index_structure=index_structure,
            re_balance_signal=re_balance_signal,
            roll_info=roll_info,
            roll_weight=roll_weight,
            prices=prices,
            level=level,
            trade_halt=trade_halt,
            **kwargs,
        )
    )
    DebugContext.print("[re_balance_index] index_structure:post", new_index_structure)
    return new_index_structure


@graph
def _pass_through(tsb: TSB[TS_SCHEMA]) -> TSB[IndexStructure]:
    return dedup(tsb.index_structure)  # Required as there seems to be an issue with a simple ref pass through


@graph
def _re_balance(
        tsb: TSB[TS_SCHEMA],
        extract_target_units_fn: Callable[[TSB[TS_SCHEMA]], NotionalUnitValues]
) -> TSB[IndexStructure]:
    # Ensure we have a valid value when we enter (This should only enter initially when we re-balance)
    end_roll = dedup(
        sample(
            (dt := (roll_info := tsb.roll_info).dt),
            default(
                gate(not_(trade_halt := tsb.trade_halt), roll_info.as_schema.end_roll),
                False
            )
        )
    )
    rebase_index = roll_info.as_schema.begin_roll
    DebugContext.print("[re_balance_index] rebase_index", rebase_index)
    DebugContext.print("[re_balance_index] end_roll", end_roll)
    DebugContext.print("[re_balance_index] re_balance_signal", re_balance_signal := tsb.re_balance_signal)
    DebugContext.print("[re_balance_index] roll_weight", roll_weight := tsb.roll_weight)

    # Compute the portfolio change
    previous_units = if_then_else(
        rebase_index,
        (current_units := (current_position := (index_structure := tsb.index_structure).current_position).units),
        index_structure.previous_units
    )
    DebugContext.print("[re_balance_index] previous_units", previous_units)

    target_units = if_then_else(  # Try and replace with switch
        rebase_index,
        extract_target_units_fn(
            tsb
        ),
        index_structure.target_units
    )
    DebugContext.print("[re_balance_index] target_units", target_units)

    # Then we need to compute the time-related weighting when we are rolling
    rolled_units = switch_(
        roll_info.roll_state,
        {
            CmpResult.LT: lambda c, p, t, w, h, e_r: if_then_else(e_r, t, c),
            CmpResult.EQ: lambda c, p, t, w, h, e_r: roll_units(c, p, t, w, h),
            CmpResult.GT: lambda c, p, t, w, h, e_r: if_then_else(h, c, t)
        },
        current_units,
        previous_units,
        target_units,
        roll_weight,
        trade_halt,
        end_roll
    )
    DebugContext.print("[re_balance_index] rolled_units", rolled_units)

    # Detect "trade" and update the current positions to reflect said trade
    traded = not_(rolled_units == current_units)
    DebugContext.print("traded", traded)
    rolled_position = switch_(
        traded,
        {
            True: lambda c_p, c_u, p, l: combine[TSB[IndexPosition]](
                units=c_u,
                level=l,
                unit_values=map_(lambda u, p: p, c_u, no_key(p))
            ),
            False: lambda c_p, c_u, p, l: c_p
        },
        current_position,
        rolled_units,
        tsb.prices,
        tsb.level
    )
    DebugContext.print("[re_balance_index] rolled_position", rolled_position)

    # Detect the end-roll and adjust as appropriate
    empty_units = const(frozendict(), NotionalUnits)

    return combine[TSB[IndexStructure]](
        current_position=rolled_position,
        previous_units=if_then_else(end_roll, empty_units, previous_units),
        target_units=if_then_else(end_roll, empty_units, target_units),
    )


@graph
def roll_units(
        current_units: NotionalUnits,
        previous_units: NotionalUnits,
        target_units: NotionalUnits,
        roll_weight: TS[float],
        roll_halted: TS[bool],
) -> NotionalUnits:
    """
    Converts the units from one contract to another.
    The ration of conversion is managed by the roll_weight.
    If we are in roll halt mode then we do not convert, but instead return the
    current units value.
    This produce a new set of current units from the combination of the previous and
    the target contracts. Roll is completed when the result matches the target units.
    """
    return switch_(
        roll_halted,
        {
            True: lambda c, p, t, w: c,
            False: lambda c, p, t, w: _roll_units(p, t, w)
        },
        current_units,
        previous_units,
        target_units,
        roll_weight
    )


@graph
def _roll_units(prev_units: NotionalUnits, target_units: NotionalUnits, weights: TS[float]) -> NotionalUnits:
    prev = map_(lambda u, w: u * w, prev_units, weights)
    target = map_(lambda u, w: u * (1.0 - w), target_units, weights)
    return map_(lambda p, t: default(p, 0.0) + default(t, 0.0), prev, target)



from dataclasses import dataclass
from datetime import date
from typing import Mapping

from hgraph import CompoundScalar, compute_node, TS, TSB

__all__ = ["IndexConfiguration", "SingleAssetIndexConfiguration", "MultiIndexConfiguration", "initial_structure_from_config"]

from hg_systematic.index.units import IndexStructure


@dataclass(frozen=True)
class IndexConfiguration(CompoundScalar):
    """

    publish_holiday_calendar: str
        The calendar to use for publishing the index.

    rounding: int
        The number of decimal places to round the published result to

    initial_level: float
        The level to start the index at.

    start_date: date
        The first date of the index. Since the level is path dependent, the start date is required.
    """
    symbol: str
    publish_holiday_calendar: str = None
    rounding: int = 8
    initial_level: float = 100.0
    start_date: date = None
    current_position: Mapping[str, float] = None
    current_position_value: Mapping[str, float] = None
    current_level: float = 100.0
    target_position: Mapping[str, float] = None
    previous_position: Mapping[str, float] = None


@dataclass(frozen=True)
class SingleAssetIndexConfiguration(IndexConfiguration):
    """
    In order to set appropriate initial conditions, the position data is available to be set.

    asset: str
        The asset symbol. Used to construct the contract name.

    initial_level: float
        Defaulted to 100.0
        If this is expected to start from a positions within the stream of index values, then the initial
        conditions for the positions tracking is also required.
    """
    asset: str = None


@dataclass(frozen=True)
class MultiIndexConfiguration(IndexConfiguration):
    indices: tuple[str, ...] = None


@compute_node
def initial_structure_from_config(config: TS[IndexConfiguration]) -> TSB[IndexStructure]:
    """
    Prepare the initial structure from the index configuration.
    This will tick once only with the values extracted from the index configuration.
    """
    config.make_passive()
    config: IndexConfiguration = config.value
    return {
        "current_position": {
            "units": {} if config.current_position is None else config.current_position,
            "unit_values": {} if config.current_position is None else config.current_position_value,
            "level": config.current_level
        },
        "previous_units": {} if config.previous_position is None else config.previous_position,
        "target_units": {} if config.target_position is None else config.target_position,
    }
